unit gfx_colorstatistics;
{***********************************************************************
Unit gfx_colorstatistics.PAS v1.5 0801
    (c) by Andreas Moser, amoser@amoser.de

    Delphi version : Delphi 4, 6 (due to user responses it seem to work with Delphi 5)

    gfx_colorstatistics is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.

    Please report bugs to:
    Andreas Moser  amoser@amoser.de

    descriptive statistic algorythms are based on or taken from the sources written by
    Earl F. Glynn, efg's computer lab

********************************************************************************}



interface
uses gfx_basedef,gfx_syswrapper;

// -----------------------------------------------------------------------------
//
//			Statistics
//
// -----------------------------------------------------------------------------

  const
    NAN = MaxDouble / 2.0;

  type
   TDescriptiveStatistics =
   class(TObject)
     protected
       FCount       :  INTEGER;
       FMaxValue    :  DOUBLE;
       FMinValue    :  DOUBLE;
       FSum         :  DOUBLE;
       FSumOfSquares:  DOUBLE;
     public
       property  Count   :  INTEGER   read FCount;
       property  MaxValue:  DOUBLE    read FMaxValue;
       property  MinValue:  DOUBLE    read FMinValue;
       property  Sum     :  DOUBLE    read FSum;
       constructor Create;
       procedure NextValue(const x:  DOUBLE);
       function MeanValue        :  DOUBLE;
       function StandardDeviation:  DOUBLE;
       procedure ResetValues;
   end;


// -----------------------------------------------------------------------------
//
//			ColorSpace
//
// -----------------------------------------------------------------------------

 type
    THistogram  = array[0..255] of INTEGER;
    TRGBStatistics =
    class(TObject)
      protected
        FStatsRed  :  TDescriptiveStatistics;
        FStatsGreen:  TDescriptiveStatistics;
        FStatsBlue :  TDescriptiveStatistics;

        function PixelCount:  INTEGER;

      public
        property    Count:  INTEGER  read PixelCount;

        constructor Create;
        destructor  Destroy;  override;
        procedure NextRGBTriple(const rgb:  TxRGBTriple);
        procedure ResetValues;
      end;

  { ---------------------------------------------------------------- }

  // Bitmap Statistics
  procedure GetBitmapStatistics(const Bitmap:  TxBitmap;
                                const ColorPlane:  TColorPlane;
                                var Statistics:  TDescriptiveStatistics);
  procedure GetHistogram(const Bitmap:  TxBitmap;
                         const ColorPlane:  TColorPlane;
                         var   Histogram:  THistogram);
  procedure DrawHistogram(const ColorPlane:  TColorPlane;
                          const Histogram:  THistogram;
                          const aCanvas:  TxCanvas;
                          const LineColor:TxColor;                          //** Added
                          const TextColor:TxColor);                         //** Added


implementation


//------------------------------------------------------------
// Statistics
//------------------------------------------------------------


// -----------------------------------------------------------------------------
//
//			Create
//
// -----------------------------------------------------------------------------
  constructor TDescriptiveStatistics.Create;
  begin
    ResetValues
  end {Create};


// -----------------------------------------------------------------------------
//
//			NextValue
//
// -----------------------------------------------------------------------------
  procedure TDescriptiveStatistics.NextValue (const x:  DOUBLE);
  begin
    INC(FCount);
    FSum := FSum + x;
    FSumOfSquares := FSumOfSquares + x*x;

    if   x > FMaxValue
    then FmaxValue := x;

    if  x < FMinValue
    then FMinValue := x
  end {NextNumber};


// -----------------------------------------------------------------------------
//
//			MeanValue
//
// -----------------------------------------------------------------------------
  function TDescriptiveStatistics.MeanValue:  DOUBLE;
  begin
    if   FCount = 0
    then RESULT := NAN
    else RESULT := FSum / FCount
  end {MeanValue};


// -----------------------------------------------------------------------------
//
//			StandardDeviation
//
// -----------------------------------------------------------------------------
  function TDescriptiveStatistics.StandardDeviation:  DOUBLE;
  begin
    if   FCount < 2
    then RESULT := NAN
    else RESULT := SQRT( (FSumOfSquares - (FSum*FSum / FCount)) /
                         (FCount - 1) )
  end {StandardDeviation};


// -----------------------------------------------------------------------------
//
//			ResetValue
//
// -----------------------------------------------------------------------------
  procedure TDescriptiveStatistics.ResetValues;
  begin
    FCount := 0;

    FMinValue :=  MaxDouble;
    FMaxValue := -MaxDouble;

    FSum          := 0.0;
    FSumOfSquares := 0.0
  end {ResetValues};

//------------------------------------------------------------
// Colorspace
//------------------------------------------------------------

// -----------------------------------------------------------------------------
//
//			Create
//
// -----------------------------------------------------------------------------
constructor TRGBStatistics.Create;
begin
    FStatsRed   := TDescriptiveStatistics.Create;
    FStatsGreen := TDescriptiveStatistics.Create;
    FStatsBlue  := TDescriptiveStatistics.Create;

    FStatsRed.ResetValues;
    FStatsGreen.ResetValues;
    FStatsBlue.ResetValues
end;


// -----------------------------------------------------------------------------
//
//			Destroy
//
// -----------------------------------------------------------------------------
destructor TRGBStatistics.Destroy;
begin
    FStatsRed.Free;
    FStatsGreen.Free;
    FStatsBlue.Free
end;


// -----------------------------------------------------------------------------
//
//			PixelCount
//
// -----------------------------------------------------------------------------
function TRGBStatistics.PixelCount;
begin
    RESULT := FStatsRed.Count
end;


// -----------------------------------------------------------------------------
//
//			NextRGBTriple
//
// -----------------------------------------------------------------------------
procedure TRGBStatistics.NextRGBTriple(const rgb:  TxRGBTriple);
begin
    with rgb DO
    begin
      FStatsRed.NextValue(rgbtRed);
      FStatsGreen.NextValue(rgbtGreen);
      FStatsBlue.NextValue(rgbtBlue);
    end
end;

// -----------------------------------------------------------------------------
//
//			ResetValues
//
// -----------------------------------------------------------------------------
procedure TRGBStatistics.ResetValues;
begin
    FStatsRed.ResetValues;
    FStatsGreen.ResetValues;
    FStatsBlue.ResetValues
end;

{==  Image Statistics  ==============================================}

// -----------------------------------------------------------------------------
//
//			GetBitmapStatistics
//
// -----------------------------------------------------------------------------
procedure GetBitmapStatistics(const Bitmap:  TxBitmap;const ColorPlane:  TColorPlane;
          var Statistics:  TDescriptiveStatistics);
var
      CMYK:  TCMYKQuad;
      HSV  :  THSVTriple;      {color coordinates}
      i      :  INTEGER;
      j      :  INTEGER;
      Row    :  pRGBarray;
      Value  :  INTEGER;
begin
    IF   Bitmap.Pixelformat <> xpf24bit
    THEN RAISE EGraphicformat.Create('GetBitmapStatistics:  ' +
               'Bitmap must be 24-bit color.');

    value := 0;  {avoid compiler warning about initialization}

    {Step through each row of image.}
    for j := Bitmap.Height-1 downto 0 DO
    begin
      Row  := pRGBarray(Bitmap.Scanline[j]);

      for i := Bitmap.Width-1 downto 0 DO
      begin
        CASE ColorPlane OF
          cpRGB,
          cpIntensity:   value := RGBIntensity(Row[i]);

          cpRed:        value := Row[i].rgbtRed;
          cpGreen:      value := Row[i].rgbtGreen;
          cpBlue:       value := Row[i].rgbtBlue;

          cpHueHSV:
            begin
              RGBtoHSV(Row[i], HSV);
              value := HSV.hsvHue
            end;

          cpSaturationHSV:
           begin
              RGBtoHSV(Row[i], HSV);
              value := HSV.hsvSaturation
            end;

          cpValue:
            begin
              RGBtoHSV(Row[i], HSV);
              value := HSV.hsvValue
            end;

          cpCyan:
            begin
              RGBtoCMYK(Row[i], CMYK);
              value := CMYK.cmykCyan
            end

        end;

        Statistics.NextValue(value)

      end

    end

end;

// -----------------------------------------------------------------------------
//
//			GetHistogram
//
// -----------------------------------------------------------------------------
procedure GetHistogram(const Bitmap    :  TxBitmap;
                         const ColorPlane:  TColorPlane;
                         var Histogram   :  THistogram);
var
      CMYK:  TCMYKQuad;
      HSV  :  THSVTriple;      {color coordinates}
      i      :  INTEGER;
      index  :  INTEGER;
      j      :  INTEGER;
      Row    :  pRGBarray;
      Row8bit:  pxBytearray;
begin
    IF   (Bitmap.Pixelformat <> xpf24bit) AND
         (Bitmap.Pixelformat <> xpf8bit)
    THEN RAISE EGraphicformat.Create('GetHistogram:  ' +
                   'Bitmap must be 8-bit or 24-bit.');

    index := 0;  {avoid compiler warning about initialization}

    for i := Low(THistogram) TO High(THistogram) DO
      Histogram[i] := 0;

    IF  Bitmap.Pixelformat = xpf8bit
    THEN begin
      IF   ColorPlane <> cp8bit
      THEN RAISE EGraphicformat.Create('GetHistogram:  ' +
               'Attempting to extract color plane from 8-bit bitmap.');

      for j := Bitmap.Height-1 downto 0 DO
      begin
        Row8bit  := pxBytearray(Bitmap.Scanline[j]);
        for i := Bitmap.Width-1 downto 0 DO
        begin
          INC (Histogram[Row8bit[i]])
        end
      end

    end
    else begin
      {Step through each row of image.}
      for j := Bitmap.Height-1 downto 0 DO
      begin
        Row  := pRGBarray(Bitmap.Scanline[j]);

        for i := Bitmap.Width-1 downto 0 DO
        begin
          CASE ColorPlane OF
            cpRGB,
            cpIntensity:   index := RGBIntensity(Row[i]);

            cpRed:        index := Row[i].rgbtRed;
            cpGreen:      index := Row[i].rgbtGreen;
            cpBlue:       index := Row[i].rgbtBlue;

            cpHueHSV:
              begin
                RGBtoHSV(Row[i], HSV);
                index := (255 * HSV.hsvHue) DIV 359
              end;

            cpSaturationHSV:
              begin
                RGBtoHSV(Row[i], HSV);
                index := HSV.hsvSaturation
              end;

            cpValue:
              begin
                RGBtoHSV(Row[i], HSV);
                index := HSV.hsvValue
              end;

            cpCyan:
              begin
                RGBtoCMYK(Row[i], CMYK);
                index := CMYK.cmykCyan
              end;

            cpMagenta:
              begin
                RGBtoCMYK(Row[i], CMYK);
                index := CMYK.cmykMagenta
              end;

            cpYellow:
              begin
                RGBtoCMYK(Row[i], CMYK);
                index := CMYK.cmykYellow
              end;

            cpBlack:
              begin
                RGBtoCMYK(Row[i], CMYK);
                index := CMYK.cmykK
              end;

          end;

          INC (Histogram[index]);
        end

      end

    end
end;


// -----------------------------------------------------------------------------
//
//			DrawHistogram
//
// -----------------------------------------------------------------------------
procedure DrawHistogram(const ColorPlane:  TColorPlane;
                          const Histogram :  THistogram;
                          const aCanvas    :  TxCanvas;
                          const LineColor:TxColor;                     //<- LineColor added 19.04.98
                          const TextColor:TxColor);                     //<- TextColor added 19.04.98

const
      MajorTickSize = 8;

var
      BarLength:  INTEGER;
      Delta    :  INTEGER;
      dheight  :  Integer;                                                  //** Added
      Color    :  TxColor;
      i        :  INTEGER;
      j        :  INTEGER;
      MaxValue :  INTEGER;
      MinValue :  Integer;                                                  //** Added
      RGBTriple:  TxRGBTRiple;
      Width    :  INTEGER;
      lMargin  :  Integer;                                                  //** Added
      HSV:      THSVTriple;
begin
    Color := xclBlack;    {avoid compiler warning about initialization}
    aCanvas.Pen.Style:=xpsSolid;

    lMargin:=44;                                                            //** Added
    dHeight   := aCanvas.ClipRect.Bottom-(4+MajorTickSize+aCanvas.TextHeight('X'));   //** Added
    Width    := aCanvas.ClipRect.Right;

    MaxValue := xMaxIntValue(Histogram);
    MinValue := xMinIntValue(Histogram);

    {for now only paint on a canvas exactly 256 pixels wide.  If
    MaxValue is zero, array was not filled in correctly and is ignored.}
    IF   (Width = 256+lMargin) AND (MaxValue > 0)
    THEN begin
      for i := Low(THistogram) TO High(THistogram) DO
      begin
        CASE ColorPlane OF
          cpRGB,
          cpIntensity:   Color := xRGB(i, i, i);

          cpRed:         Color := xRGB(i, 0, 0);
          cpGreen:       Color := xRGB(0, i, 0);
          cpBlue:        Color := xRGB(0, 0, i);
          cpHueHSV:
            begin
              HSV.hsvHue:=(i*359) DIV 255;
              HSV.hsvSaturation:=255;
              HSV.hsvValue:=255;
              HSVtoRGB(HSV, RGBTriple);
              with RGBTriple DO
                Color := xRGB(rgbtRed, rgbtGreen, rgbtBlue)
            end;
          cpSaturationHSV:
                         Color := xclBlack;
          cpValue:       Color := xRGB(i, i, i);

        end;

        aCanvas.Pen.Color := Color;
        BarLength := ROUND(dHeight*Histogram[i] / MaxValue);
        aCanvas.MoveTo(i+lMargin, dHeight-1);
        aCanvas.LineTo(i+lMargin, dHeight-1-BarLength)
//          aCanvas.Pixels[i+lMargin, dHeight-1-BarLength]:=Color;
      end;

      aCanvas.Pen.Color := LineColor;
      {Vertical Lines for visual estimation}
      for i := 0 TO 25 DO
      begin
        aCanvas.MoveTo((10*i)+lMargin, dHeight-1);
        IF   i MOD 5 = 0
        THEN Delta := MajorTickSize
        else Delta := MajorTickSize DIV 2;
        aCanvas.LineTo((10*i)+lMargin, dHeight+1+Delta);
      end;

      {Horizontal Lines}
      for j := 0 TO 4 DO
      begin
        aCanvas.MoveTo(0+lMargin, j*dHeight DIV 5);
        aCanvas.LineTo(Width-1, j*dHeight DIV 5);
      end;

      //** Added +
      aCanvas.Font.Color:=TextColor;
      aCanvas.Brush.Style := xbsClear;
      aCanvas.TextOut(lMargin-aCanvas.TextWidth(xIntToStr(MaxValue))-2, 0, xIntToStr(MaxValue));
      aCanvas.TextOut(lMargin-aCanvas.TextWidth(xIntToStr(MinValue))-2, dHeight, xIntToStr(MinValue));
      //** Added -

      aCanvas.TextOut(2+lMargin, dHeight+aCanvas.TextHeight('X') - MajorTickSize+3, '0');
      aCanvas.TextOut(Width-aCanvas.TextWidth('250 '),
                     dHeight+aCanvas.TextHeight('X') - MajorTickSize+3, '250')
    end
end;


end.
